@@ -28,6 +28,7 @@
 #include <string.h>
 #include <stdlib.h>
 
+#include "galias.h"
 #include "glib.h"
 #include "gprintfint.h"
 #include "gthreadinit.h"
@@ -90,7 +91,7 @@ try_to_aliases (const char **to_aliases,
   return FALSE;
 }
 
-extern const char **_g_charset_get_aliases (const char *canonical_name);
+extern const char **_g_charset_get_aliases (const char *canonical_name) G_GNUC_INTERNAL;
 
 /**
  * g_iconv_open:
@@ -403,8 +404,8 @@ open_converter (const gchar *to_codeset,
 		 from_codeset, to_codeset);
   else
     g_set_error (error, G_CONVERT_ERROR, G_CONVERT_ERROR_FAILED,
-		 _("Could not open converter from '%s' to '%s': %s"),
-		 from_codeset, to_codeset, g_strerror (errno));
+		 _("Could not open converter from '%s' to '%s'"),
+		 from_codeset, to_codeset);
   
   return cd;
 }
@@ -912,7 +913,8 @@ strdup_len (const gchar *string,
 
 /**
  * g_locale_to_utf8:
- * @opsysstring:   a string in the encoding of the current locale
+ * @opsysstring:   a string in the encoding of the current locale. On Windows
+ *                 this means the system codepage.
  * @len:           the length of the string, or -1 if the string is
  *                 nul-terminated.
  * @bytes_read:    location to store the number of bytes in the
@@ -992,34 +994,172 @@ g_locale_from_utf8 (const gchar *utf8string,
 
 #ifndef G_PLATFORM_WIN32
 
-static gboolean
-have_broken_filenames (void)
-{
-  static gboolean initialized = FALSE;
-  static gboolean broken;
-  
-  if (initialized)
-    return broken;
+typedef struct _GFilenameCharsetCache GFilenameCharsetCache;
 
-  broken = (getenv ("G_BROKEN_FILENAMES") != NULL);
-  
-  initialized = TRUE;
-  
-  return broken;
+struct _GFilenameCharsetCache {
+  gboolean is_utf8;
+  gchar *charset;
+  gchar **filename_charsets;
+};
+
+static void
+filename_charset_cache_free (gpointer data)
+{
+  GFilenameCharsetCache *cache = data;
+  g_free (cache->charset);
+  g_strfreev (cache->filename_charsets);
+  g_free (cache);
 }
+
+/**
+ * g_get_filename_charsets:
+ * @charsets: return location for the %NULL-terminated list of encoding names
+ *
+ * Determines the preferred character sets used for filenames.
+ * The first character set from the @charsets is the filename encoding, the
+ * subsequent character sets are used when trying to generate a displayable
+ * representation of a filename, see g_filename_display_name().
+ *
+ * On Unix, the character sets are determined by consulting the
+ * environment variables <envar>G_FILENAME_ENCODING</envar> and
+ * <envar>G_BROKEN_FILENAMES</envar>. On Windows, the character set
+ * used in the GLib API is always UTF-8 and said environment variables
+ * have no effect.
+ *
+ * <envar>G_FILENAME_ENCODING</envar> may be set to a comma-separated list 
+ * of character set names. The special token "@locale" is taken to mean the 
+ * character set for the current locale. If <envar>G_FILENAME_ENCODING</envar> 
+ * is not set, but <envar>G_BROKEN_FILENAMES</envar> is, the character set of 
+ * the current locale is taken as the filename encoding. If neither environment
+ * variable is set, UTF-8 is taken as the filename encoding, but the character
+ * set of the current locale is also put in the list of encodings.
+ *
+ * The returned @charsets belong to GLib and must not be freed.
+ *
+ * Note that on Unix, regardless of the locale character set or
+ * <envar>G_FILENAME_ENCODING</envar> value, the actual file names present on a
+ * system might be in any random encoding or just gibberish.
+ *
+ * Return value: %TRUE if the filename encoding is UTF-8.
+ * 
+ * Since: 2.6
+ */
+gboolean
+g_get_filename_charsets (G_CONST_RETURN gchar ***filename_charsets)
+{
+  static GStaticPrivate cache_private = G_STATIC_PRIVATE_INIT;
+  GFilenameCharsetCache *cache = g_static_private_get (&cache_private);
+  const gchar *charset;
+
+  if (!cache)
+    {
+      cache = g_new0 (GFilenameCharsetCache, 1);
+      g_static_private_set (&cache_private, cache, filename_charset_cache_free);
+    }
+
+  g_get_charset (&charset);
+
+  if (!(cache->charset && strcmp (cache->charset, charset) == 0))
+    {
+      const gchar *new_charset;
+      gchar *p;
+      gint i;
+
+      g_free (cache->charset);
+      g_strfreev (cache->filename_charsets);
+      cache->charset = g_strdup (charset);
+      
+      p = getenv ("G_FILENAME_ENCODING");
+      if (p != NULL) 
+	{
+	  cache->filename_charsets = g_strsplit (p, ",", 0);
+	  cache->is_utf8 = (strcmp (cache->filename_charsets[0], "UTF-8") == 0);
+
+	  for (i = 0; cache->filename_charsets[i]; i++)
+	    {
+	      if (strcmp ("@locale", cache->filename_charsets[i]) == 0)
+		{
+		  g_get_charset (&new_charset);
+		  g_free (cache->filename_charsets[i]);
+		  cache->filename_charsets[i] = g_strdup (new_charset);
+		}
+	    }
+	}
+      else if (getenv ("G_BROKEN_FILENAMES") != NULL)
+	{
+	  cache->filename_charsets = g_new0 (gchar *, 2);
+	  cache->is_utf8 = g_get_charset (&new_charset);
+	  cache->filename_charsets[0] = g_strdup (new_charset);
+	}
+      else 
+	{
+	  cache->filename_charsets = g_new0 (gchar *, 3);
+	  cache->is_utf8 = TRUE;
+	  cache->filename_charsets[0] = g_strdup ("UTF-8");
+	  if (!g_get_charset (&new_charset))
+	    cache->filename_charsets[1] = g_strdup (new_charset);
+	}
+    }
+
+  if (filename_charsets)
+    *filename_charsets = (const gchar **)cache->filename_charsets;
+
+  return cache->is_utf8;
+}
+
 #else /* G_PLATFORM_WIN32 */
 
-#define have_broken_filenames() TRUE
+gboolean
+g_get_filename_charsets (G_CONST_RETURN gchar ***filename_charsets) 
+{
+  static gchar *charsets[] = {
+    "UTF-8",
+    NULL
+  };
+
+#ifdef G_OS_WIN32
+  /* On Windows GLib pretends that the filename charset is UTF-8 */
+  if (filename_charsets)
+    *filename_charsets = charsets;
+
+  return TRUE;
+#else
+  gboolean result;
+
+  /* Cygwin works like before */
+  result = g_get_charset (&(charsets[0]));
+
+  if (filename_charsets)
+    *filename_charsets = charsets;
+
+  return result;
+#endif
+}
 
 #endif /* G_PLATFORM_WIN32 */
 
+static gboolean
+get_filename_charset (const gchar **filename_charset)
+{
+  const gchar **charsets;
+  gboolean is_utf8;
+  
+  is_utf8 = g_get_filename_charsets (&charsets);
+
+  if (filename_charset)
+    *filename_charset = charsets[0];
+  
+  return is_utf8;
+}
+
 /* This is called from g_thread_init(). It's used to
  * initialize some static data in a threadsafe way.
  */
 void 
 _g_convert_thread_init (void)
 {
-  (void)have_broken_filenames ();
+  const gchar **dummy;
+  (void) g_get_filename_charsets (&dummy);
 }
 
 /**
@@ -1040,7 +1180,7 @@ _g_convert_thread_init (void)
  * @error:         location to store the error occuring, or %NULL to ignore
  *                 errors. Any of the errors in #GConvertError may occur.
  * 
- * Converts a string which is in the encoding used for filenames
+ * Converts a string which is in the encoding used by GLib for filenames
  * into a UTF-8 string.
  * 
  * Return value: The converted string, or %NULL on an error.
@@ -1052,14 +1192,39 @@ g_filename_to_utf8 (const gchar *opsysstring,
 		    gsize       *bytes_written,
 		    GError     **error)
 {
-  if (have_broken_filenames ())
-    return g_locale_to_utf8 (opsysstring, len,
-			     bytes_read, bytes_written,
-			     error);
-  else
+  const gchar *charset;
+
+  if (get_filename_charset (&charset))
     return strdup_len (opsysstring, len, bytes_read, bytes_written, error);
+  else
+    return g_convert (opsysstring, len, 
+		      "UTF-8", charset, bytes_read, bytes_written, error);
 }
 
+#ifdef G_OS_WIN32
+
+#undef g_filename_to_utf8
+
+/* Binary compatibility version. Not for newly compiled code. */
+
+gchar*
+g_filename_to_utf8 (const gchar *opsysstring, 
+		    gssize       len,           
+		    gsize       *bytes_read,   
+		    gsize       *bytes_written,
+		    GError     **error)
+{
+  const gchar *charset;
+
+  if (g_get_charset (&charset))
+    return strdup_len (opsysstring, len, bytes_read, bytes_written, error);
+  else
+    return g_convert (opsysstring, len, 
+		      "UTF-8", charset, bytes_read, bytes_written, error);
+}
+
+#endif
+
 /**
  * g_filename_from_utf8:
  * @utf8string:    a UTF-8 encoded string.
@@ -1089,14 +1254,39 @@ g_filename_from_utf8 (const gchar *utf8string,
 		      gsize       *bytes_written,
 		      GError     **error)
 {
-  if (have_broken_filenames ())
-    return g_locale_from_utf8 (utf8string, len,
-			       bytes_read, bytes_written,
-			       error);
-  else
+  const gchar *charset;
+
+  if (get_filename_charset (&charset))
     return strdup_len (utf8string, len, bytes_read, bytes_written, error);
+  else
+    return g_convert (utf8string, len,
+		      charset, "UTF-8", bytes_read, bytes_written, error);
 }
 
+#ifdef G_OS_WIN32
+
+#undef g_filename_from_utf8
+
+/* Binary compatibility version. Not for newly compiled code. */
+
+gchar*
+g_filename_from_utf8 (const gchar *utf8string,
+		      gssize       len,            
+		      gsize       *bytes_read,    
+		      gsize       *bytes_written,
+		      GError     **error)
+{
+  const gchar *charset;
+
+  if (g_get_charset (&charset))
+    return strdup_len (utf8string, len, bytes_read, bytes_written, error);
+  else
+    return g_convert (utf8string, len,
+		      charset, "UTF-8", bytes_read, bytes_written, error);
+}
+
+#endif
+
 /* Test of haystack has the needle prefix, comparing case
  * insensitive. haystack may be UTF-8, but needle must
  * contain only ascii. */
@@ -1315,7 +1505,7 @@ g_unescape_uri_string (const char *escaped,
   g_assert (out - result <= len);
   *out = '\0';
 
-  if (in != in_end || !g_utf8_validate (result, -1, NULL))
+  if (in != in_end)
     {
       g_free (result);
       return NULL;
@@ -1325,15 +1515,15 @@ g_unescape_uri_string (const char *escaped,
 }
 
 static gboolean
-is_escalphanum (gunichar c)
+is_asciialphanum (gunichar c)
 {
-  return c > 0x7F || g_ascii_isalnum (c);
+  return c <= 0x7F && g_ascii_isalnum (c);
 }
 
 static gboolean
-is_escalpha (gunichar c)
+is_asciialpha (gunichar c)
 {
-  return c > 0x7F || g_ascii_isalpha (c);
+  return c <= 0x7F && g_ascii_isalpha (c);
 }
 
 /* allows an empty string */
@@ -1351,7 +1541,7 @@ hostname_validate (const char *hostname)
       /* read in a label */
       c = g_utf8_get_char (p);
       p = g_utf8_next_char (p);
-      if (!is_escalphanum (c))
+      if (!is_asciialphanum (c))
 	return FALSE;
       first_char = c;
       do
@@ -1360,13 +1550,13 @@ hostname_validate (const char *hostname)
 	  c = g_utf8_get_char (p);
 	  p = g_utf8_next_char (p);
 	}
-      while (is_escalphanum (c) || c == '-');
+      while (is_asciialphanum (c) || c == '-');
       if (last_char == '-')
 	return FALSE;
       
       /* if that was the last label, check that it was a toplabel */
       if (c == '\0' || (c == '.' && *p == '\0'))
-	return is_escalpha (first_char);
+	return is_asciialpha (first_char);
     }
   while (c == '.');
   return FALSE;
@@ -1374,14 +1564,14 @@ hostname_validate (const char *hostname)
 
 /**
  * g_filename_from_uri:
- * @uri: a uri describing a filename (escaped, encoded in UTF-8).
+ * @uri: a uri describing a filename (escaped, encoded in ASCII).
  * @hostname: Location to store hostname for the URI, or %NULL.
  *            If there is no hostname in the URI, %NULL will be
  *            stored in this location.
  * @error: location to store the error occuring, or %NULL to ignore
  *         errors. Any of the errors in #GConvertError may occur.
  * 
- * Converts an escaped UTF-8 encoded URI to a local filename in the
+ * Converts an escaped ASCII-encoded URI to a local filename in the
  * encoding used for filenames. 
  * 
  * Return value: a newly-allocated string holding the resulting
@@ -1408,7 +1598,7 @@ g_filename_from_uri (const gchar *uri,
   if (!has_case_prefix (uri, "file:/"))
     {
       g_set_error (error, G_CONVERT_ERROR, G_CONVERT_ERROR_BAD_URI,
-		   _("The URI '%s' is not an absolute URI using the file scheme"),
+		   _("The URI '%s' is not an absolute URI using the \"file\" scheme"),
 		   uri);
       return NULL;
     }
@@ -1501,33 +1691,56 @@ g_filename_from_uri (const gchar *uri,
 	}
     }
 #endif
-  
-  result = g_filename_from_utf8 (filename + offs, -1, NULL, NULL, error);
+
+  result = g_strdup (filename + offs);
   g_free (filename);
-  
+
   return result;
 }
 
+#ifdef G_OS_WIN32
+
+#undef g_filename_from_uri
+
+gchar *
+g_filename_from_uri (const gchar *uri,
+		     gchar      **hostname,
+		     GError     **error)
+{
+  gchar *utf8_filename;
+  gchar *retval = NULL;
+
+  utf8_filename = g_filename_from_uri_utf8 (uri, hostname, error);
+  if (utf8_filename)
+    {
+      retval = g_locale_from_utf8 (utf8_filename, -1, NULL, NULL, error);
+      g_free (utf8_filename);
+    }
+  return retval;
+}
+
+#endif
+
 /**
  * g_filename_to_uri:
- * @filename: an absolute filename specified in the encoding
- *            used for filenames by the operating system.
+ * @filename: an absolute filename specified in the GLib file name encoding,
+ *            which is the on-disk file name bytes on Unix, and UTF-8 on 
+ *            Windows
  * @hostname: A UTF-8 encoded hostname, or %NULL for none.
  * @error: location to store the error occuring, or %NULL to ignore
  *         errors. Any of the errors in #GConvertError may occur.
  * 
- * Converts an absolute filename to an escaped UTF-8 encoded URI.
+ * Converts an absolute filename to an escaped ASCII-encoded URI.
  * 
  * Return value: a newly-allocated string holding the resulting
  *               URI, or %NULL on an error.
  **/
 gchar *
-g_filename_to_uri   (const gchar *filename,
-		     const gchar *hostname,
-		     GError     **error)
+g_filename_to_uri (const gchar *filename,
+		   const gchar *hostname,
+		   GError     **error)
 {
   char *escaped_uri;
-  char *utf8_filename;
 
   g_return_val_if_fail (filename != NULL, NULL);
 
@@ -1548,20 +1761,245 @@ g_filename_to_uri   (const gchar *filename,
       return NULL;
     }
   
-  utf8_filename = g_filename_to_utf8 (filename, -1, NULL, NULL, error);
-  if (utf8_filename == NULL)
-    return NULL;
-  
 #ifdef G_OS_WIN32
   /* Don't use localhost unnecessarily */
   if (hostname && g_ascii_strcasecmp (hostname, "localhost") == 0)
     hostname = NULL;
 #endif
 
-  escaped_uri = g_escape_file_uri (hostname,
-				   utf8_filename);
-  g_free (utf8_filename);
-  
+  escaped_uri = g_escape_file_uri (hostname, filename);
+
   return escaped_uri;
 }
 
+#ifdef G_OS_WIN32
+
+#undef g_filename_to_uri
+
+gchar *
+g_filename_to_uri (const gchar *filename,
+		   const gchar *hostname,
+		   GError     **error)
+{
+  gchar *utf8_filename;
+  gchar *retval = NULL;
+
+  utf8_filename = g_locale_to_utf8 (filename, -1, NULL, NULL, error);
+
+  if (utf8_filename)
+    {
+      retval = g_filename_to_uri_utf8 (utf8_filename, hostname, error);
+      g_free (utf8_filename);
+    }
+
+  return retval;
+}
+
+#endif
+
+/**
+ * g_uri_list_extract_uris:
+ * @uri_list: an URI list 
+ *
+ * Splits an URI list conforming to the text/uri-list
+ * mime type defined in RFC 2483 into individual URIs,
+ * discarding any comments. The URIs are not validated.
+ *
+ * Returns: a newly allocated %NULL-terminated list of
+ *   strings holding the individual URIs. The array should
+ *   be freed with g_strfreev().
+ *
+ * Since: 2.6
+ */
+gchar **
+g_uri_list_extract_uris (const gchar *uri_list)
+{
+  GSList *uris, *u;
+  const gchar *p, *q;
+  gchar **result;
+  gint n_uris = 0;
+
+  uris = NULL;
+
+  p = uri_list;
+
+  /* We don't actually try to validate the URI according to RFC
+   * 2396, or even check for allowed characters - we just ignore
+   * comments and trim whitespace off the ends.  We also
+   * allow LF delimination as well as the specified CRLF.
+   *
+   * We do allow comments like specified in RFC 2483.
+   */
+  while (p)
+    {
+      if (*p != '#')
+	{
+	  while (g_ascii_isspace (*p))
+	    p++;
+
+	  q = p;
+	  while (*q && (*q != '\n') && (*q != '\r'))
+	    q++;
+
+	  if (q > p)
+	    {
+	      q--;
+	      while (q > p && g_ascii_isspace (*q))
+		q--;
+
+	      if (q > p)
+		{
+		  uris = g_slist_prepend (uris, g_strndup (p, q - p + 1));
+		  n_uris++;
+		}
+	    }
+	}
+      p = strchr (p, '\n');
+      if (p)
+	p++;
+    }
+
+  result = g_new (gchar *, n_uris + 1);
+
+  result[n_uris--] = NULL;
+  for (u = uris; u; u = u->next)
+    result[n_uris--] = u->data;
+
+  g_slist_free (uris);
+
+  return result;
+}
+
+static gchar *
+make_valid_utf8 (const gchar *name)
+{
+  GString *string;
+  const gchar *remainder, *invalid;
+  gint remaining_bytes, valid_bytes;
+  
+  string = NULL;
+  remainder = name;
+  remaining_bytes = strlen (name);
+  
+  while (remaining_bytes != 0) 
+    {
+      if (g_utf8_validate (remainder, remaining_bytes, &invalid)) 
+	break;
+      valid_bytes = invalid - remainder;
+    
+      if (string == NULL) 
+	string = g_string_sized_new (remaining_bytes);
+
+      g_string_append_len (string, remainder, valid_bytes);
+      g_string_append_c (string, '?');
+      
+      remaining_bytes -= valid_bytes + 1;
+      remainder = invalid + 1;
+    }
+  
+  if (string == NULL)
+    return g_strdup (name);
+  
+  g_string_append (string, remainder);
+  g_string_append (string, " (invalid encoding)");
+
+  g_assert (g_utf8_validate (string->str, -1, NULL));
+  
+  return g_string_free (string, FALSE);
+}
+
+/**
+ * g_filename_display_basename:
+ * @filename: an absolute pathname in the GLib file name encoding
+ *
+ * Returns the display basename for the particular filename, guaranteed
+ * to be valid UTF-8. The display name might not be identical to the filename,
+ * for instance there might be problems converting it to UTF-8, and some files
+ * can be translated in the display
+ *
+ * You must pass the whole absolute pathname to this functions so that
+ * translation of well known locations can be done.
+ *
+ * This function is preferred over g_filename_display_name() if you know the
+ * whole path, as it allows translation.
+ *
+ * Return value: a newly allocated string containing
+ *   a rendition of the basename of the filename in valid UTF-8
+ *
+ * Since: 2.6
+ **/
+gchar *
+g_filename_display_basename (const gchar *filename)
+{
+  char *basename;
+  char *display_name;
+
+  g_return_val_if_fail (filename != NULL, NULL);
+  
+  basename = g_path_get_basename (filename);
+  display_name = g_filename_display_name (basename);
+  g_free (basename);
+  return display_name;
+}
+
+/**
+ * g_filename_display_name:
+ * @filename: a pathname hopefully in the GLib file name encoding
+ * 
+ * Converts a filename into a valid UTF-8 string. The 
+ * conversion is not necessarily reversible, so you 
+ * should keep the original around and use the return
+ * value of this function only for display purposes.
+ * Unlike g_filename_to_utf8(), the result is guaranteed 
+ * to be non-NULL even if the filename actually isn't in the GLib
+ * file name encoding.
+ *
+ * If you know the whole pathname of the file you should use
+ * g_filename_display_basename(), since that allows location-based
+ * translation of filenames.
+ *
+ * Return value: a newly allocated string containing
+ *   a rendition of the filename in valid UTF-8
+ *
+ * Since: 2.6
+ **/
+gchar *
+g_filename_display_name (const gchar *filename)
+{
+  gint i;
+  const gchar **charsets;
+  gchar *display_name = NULL;
+  gboolean is_utf8;
+ 
+  is_utf8 = g_get_filename_charsets (&charsets);
+
+  if (is_utf8)
+    {
+      if (g_utf8_validate (filename, -1, NULL))
+	display_name = g_strdup (filename);
+    }
+  
+  if (!display_name)
+    {
+      /* Try to convert from the filename charsets to UTF-8.
+       * Skip the first charset if it is UTF-8.
+       */
+      for (i = is_utf8 ? 1 : 0; charsets[i]; i++)
+	{
+	  display_name = g_convert (filename, -1, "UTF-8", charsets[i], 
+				    NULL, NULL, NULL);
+
+	  if (display_name)
+	    break;
+	}
+    }
+  
+  /* if all conversions failed, we replace invalid UTF-8
+   * by a question mark
+   */
+  if (!display_name) 
+    display_name = make_valid_utf8 (filename);
+
+  return display_name;
+}
+
